//------------------------------------------------------------------------------
//  :	main.c 
// :	   ATtiny2313,    
//     . 
//   :
// 1.     ,    
//    
// 2.    (   ""/"")  
//      ( )
// 3.       USART
// 4.        
//
//   :
//        : 1 - 1000  ( - 0.1 )
//          0 - 100%  
//                 1   
//                10    
//          USART    
//                 - 8 , 2 -   
//                 250 /   16 
//      : 2  8  - , 
//                 HD44780
//
// :	ATtiny2313
//  : 16
// :		2.3
//
// :		 .., ku_@mail.ru
//
//  .., 2010
//------------------------------------------------------------------------------

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

#define ENCODER 1

#define RS PB2
#define E PB3
#define STROBE PB1
#define PAUSE 10      //        /
#define lmax 100      // lmax + 1 -   
#define lpc 1         // lmax*lpc -   

//  
void lcd_cmd(unsigned char);
void lcd_dat(unsigned char);
void ini_cmd(unsigned char);
void lcd_init(void);
void lcd_str(unsigned char, char*);

char fs[] = {0x20, 0x20, 0x20, 0x2E, 0x20, 0xBA, 0xA1, 0xE5, 0x00},
     ps[] = {0x20, 0x20, 0x20, 0x25, 0x20, 0x20, 0x20, 0x20, 0x00};
unsigned char lev = 0;  //  

//----------------  USART ----------------
void usart_init(unsigned int ubrr) {
  UBRRH = (unsigned char)(ubrr >> 8);
  UBRRL = (unsigned char)ubrr;
  UCSRB = _BV(TXEN);                //  
  UCSRC = _BV(UCSZ0) | _BV(UCSZ1) | _BV(USBS); //  8  
                                               // 2 -
}
//--------------------   ------------------
int usart_sb(unsigned char b) {
  if(UCSRA & _BV(UDRE)){
    UDR = b;
    return 1;
  }
  return 0;
}
//----------------     ----------------
void lcd_cmd(unsigned char p) { 	// p -  
	PORTB = (p & 0b11110000) | 
          (PORTB & 0b00001111);	    //    -   PB4-PB7
  PORTB &= ~_BV(RS);                // RS = 0
	PORTB |= _BV(E); _delay_us(100);	// E = 1
	PORTB &= ~_BV(E);	_delay_us(500); // E -> 0 -     

	PORTB = ((p << 4) & 0b11110000) |
          (PORTB & 0b00001111);	    //    -    PB4-PB7
  PORTB &= ~_BV(RS);                // RS = 0
	PORTB |= _BV(E); _delay_us(100);  // E = 1
	PORTB &= ~_BV(E);	_delay_us(500); // E -> 0 -     
}
//----------------     ----------------
void lcd_dat(unsigned char p) {		// p -  
	PORTB = (p & 0b11110000) | _BV(RS) |
          (PORTB & 0b00001111);	    //    -   PB4-PB7
  PORTB |= _BV(RS);                // RS = 1
	PORTB |= _BV(E); _delay_us(100);	// E = 1
	PORTB &= ~_BV(E);	_delay_us(500);	// E -> 0 -     

	PORTB = ((p<<4) & 0b11110000) | _BV(RS) |
          (PORTB & 0b00001111);	    //    -   PB4-PB7
  PORTB |= _BV(RS);                // RS = 1
	PORTB |= _BV(E); _delay_us(100);	// E = 1
	PORTB &= ~_BV(E); _delay_us(500);	// E -> 0 -     
}
//----------------   ----------------
void ini_cmd(unsigned char p) {   // p -  ,    
	PORTB = (p & 0b11110000) |
          (PORTB & 0b00001111);	          //    -   PB4-PB7
	PORTB &= ~_BV(RS);			          // RS = 0
	PORTB |= _BV(E); _delay_us(100);  // E = 1
	PORTB &= ~_BV(E); _delay_us(500); //     
}
//----------------   ----------------
void lcd_init() {
	_delay_ms(20); 
	ini_cmd(0x30); _delay_ms(5);
	ini_cmd(0x30); _delay_us(200);
	ini_cmd(0x30); _delay_us(100);
	ini_cmd(0x20); _delay_us(100);
	lcd_cmd(0x28); _delay_us(100);
	lcd_cmd(0x08); _delay_us(100);
	lcd_cmd(0x01); _delay_us(2000);
	lcd_cmd(0x06); _delay_us(100);
	lcd_cmd(0x0D); _delay_us(100);
}
//----------------     ----------------
void lcd_str(unsigned char l, char* s) {
	if(l == 0) //    
		lcd_cmd(0x80);
	else      //    ...
		lcd_cmd(0xC0);
	_delay_us(100);
	
	unsigned char i = 0;
	while(s[i] != 0x00)
		lcd_dat(s[i++]);
}

#ifdef ENCODER
//------     ------------
SIGNAL(SIG_INTERRUPT0){
  if(MCUCR & _BV(ISC00)){          //    0->1)

    if(PIND & _BV(PD3)){             //     B 1 -   /
      if(lev < lmax)
        lev++;
      }
    else{                           //   /
      if(lev > 0)
        lev--;
    }
    usart_sb(lev);                  //  lev  
    _delay_ms(PAUSE);              // 
    MCUCR &= ~_BV(ISC00);          //    1->0   
  }
  else{                            //   (  1->0)
    _delay_ms(PAUSE);                 // 
    MCUCR |= _BV(ISC00);           //    0->1   
  }
  EIFR = 0xFF;                     //     
}
#else
//------     "+" ----------
SIGNAL(SIG_INTERRUPT0){
  if(MCUCR & _BV(ISC00)){          // -   (  0->1)
    lev += ((lev < lmax)? 10 : 0);
    usart_sb(lev);                  //  lev  
    _delay_ms(PAUSE);                 // 
    MCUCR &= ~_BV(ISC00);          //    1->0   
  }
  else{                            //   (  1->0)
    _delay_ms(PAUSE);                 // 
    MCUCR |= _BV(ISC00);           //    0->1   
  }
  EIFR = 0xFF;                     //     
}
//------     "-" ----------
SIGNAL(SIG_INTERRUPT1){
  if(MCUCR & _BV(ISC10)){          // -   (  0->1)
    lev -= ((lev > 0)? 10 : 0);
    usart_sb(lev);                  //  lev  
    _delay_ms(PAUSE);                 // 
    MCUCR &= ~_BV(ISC10);          //    1->0   
  }
  else{                            //   (  1->0)
    _delay_ms(PAUSE);                 // 
    MCUCR |= _BV(ISC10);           //    0->1   
  }
  EIFR = 0xFF;                     //     
}
#endif
//----------------   ----------------
int main (void) {
	int f;  // 
	unsigned char w, dd, d0, d1, d2; 

	DDRB = 0xFF; PORTB = 0xFF;       //   B -  .  PB2  PB3   1

  //    - 1 (Normal)   
	TCCR1B = 0b00000111;             // CS12, cs11  CS10  1, ..    T1,
                                   //  -  
  DDRD = 0b01010011;               //  PD2, PD3  PD5 - ,  - 
                                   // PD2 -  '+', PD3 -  '-',
                                   // PD5 -   
  PORTD |= 0b00101100;             //     PD2, PD3  PD5
  //          INT0  INT1
#ifdef ENCODER
  GIMSK |= _BV(INT0);	             //      Int0
	MCUCR = _BV(ISC01);	             //     
#else
  GIMSK |= (_BV(INT0) | _BV(INT1));	  //    Int0  Int1
	MCUCR = (_BV(ISC01) | _BV(ISC11));	//      
#endif
	lcd_init();                      //  
	lcd_cmd(0x0C); _delay_us(100);
	
	usart_init(3);                   // 250 /   16
	      
  sei();
	while(1) {
    //  .     ,  ...
    //     -  ,   
    // - TIMER1  10 .  ,  
    //   delay.
    cli();                         //     
    TCNT1 = 0;                     //  
    _delay_us(10000);              //  10 
    f = TCNT1;                     //       10
    sei();
    //       
    if(f < 10000){                 //   1 
      dd = f % 10;                 //  1   
      f = f/10;                    //    
      d0 = f % 10;                 //    
      d1 = (f % 100)/10;           // -"-
      d2 = f/100;                  // -"-
      fs[0] = ((d2 == 0)? 0x20 : (0x30 + d2)); //   0 
      fs[1] = ((d2 == 0) && (d1 == 0))? 0x20 : (0x30 + d1); // ...
      fs[2] = 0x30 + d0;
      fs[4] = 0x30 + dd;
    } else {                      // ,     :-)
      fs[0] = '*'; fs[1] = '*'; fs[2] = '*'; fs[4] = '*'; 
    }
    //     lev  
    w = lev*lpc;
    d0 = w % 10;                //   lev
    d1 = (w % 100)/10;          // -"-
    d2 = w/100;                 // -"-
    ps[0] = ((d2 == 0)? 0x20 : (0x30 + d2)); //   0 
    ps[1] = ((d2 == 0) && (d1 == 0))? 0x20 : (0x30 + d1); // ...
    ps[2] = 0x30 + d0;
    lcd_str(0, fs);                //    
    lcd_str(1, ps);
    _delay_ms(250);                //  ...	
  }
	return 0;
}
